import numpy as np
from qtpy.QtCore import Signal
from pymodaq.utils.daq_utils import ThreadCommand
from pymodaq.utils.data import DataFromPlugins, Axis, DataToExport, DataRaw
from pymodaq.control_modules.viewer_utility_classes import DAQ_Viewer_base, \
    comon_parameters, main
from pymodaq.utils.parameter import Parameter
from pymodaq_plugins_avantes.hardware.AvaSpec_ULS2048CL_EVO_Controller \
    import AvantesController


class DAQ_1DViewer_Avantes(DAQ_Viewer_base):
    """ Avantes Spectrometer Instrument plugin class for a 1D viewer.
    Besides acquiring spectral data, the Avantes device may control ten digital
    output lines. They are exposed to PyMoDAQ as parameters. Acquisition on
    digital and analog input lines is not yet supported.
    """

    # define controller type for easy autocompletion
    controller_type = AvantesController

    params = comon_parameters+[
        {'title': 'Integration time:', 'name': 'integration_time',
         'type': 'float', 'min': 0.001, 'value': 0.1,
         'tip': 'Integration time in seconds'},
    ] + [ {'title': 'Output %d:' % (i + 1), 'name': 'output_%d' % (i + 1),
           'type': 'led_push', 'value': False,
           'tip': 'Logic level on putput %d' % (i + 1) } \
          for i in range(10)
         ]

    def ini_attributes(self):
        self.controller: self.controller_type = None
        self.x_axis = None

    def commit_settings(self, param: Parameter):
        if param.name() == "integration_time":
            self.controller.set_integration_time(param.value() * 1000)
        elif param.name()[:7] == 'output_':
            # Note: digital outputs are not really parameters. However and
            # for the time being, this seems to come closest to PyMoDAQ's
            # functionallity.
            self.controller.set_digital_output(int(param.name()[7:]),
                                               param.value())

    def ini_detector(self, controller=None):
        """Detector communication initialization

        Parameters
        ----------
        controller: (object)
            custom object of a PyMoDAQ plugin (Slave case). None if only 
            one actuator/detector by controller
            (Master case)

        Returns
        -------
        info: str
        initialized: bool
            False if initialization failed otherwise True
        """

        self.ini_detector_init(slave_controller=controller)

        if self.is_master:
            self.controller = self.controller_type()
            if self.controller.open_communication():
                info = "Avantes Spectro initilialized"
            else:
                info = "No Avantes Spectro detected"
                return info, False

            wavelengths = self.controller.wavelengths
            self.x_axis = Axis(label='Wavelength', units='nm',
                                data=wavelengths, index=0)
            dfp = DataFromPlugins(name='Avantes',
                                  data=[np.zeros(len(wavelengths))],
                                  dim='Data1D', axes=[self.x_axis],
                                  labels=['Avantes-Signal'])
            self.dte_signal_temp.emit(DataToExport(name='Avantes', data=[dfp]))
            self.controller.set_default_config()
        else:
            self.controller = controller

        initialized = True

        return info, initialized

    def close(self):
        """Terminate the communication protocol"""

        self.controller.close_communication()
        initialized = False

        return initialized

    def grab_data(self, Naverage=1, **kwargs):
        """Start grabbing from the detector
        Use a synchrone acquisition (blocking function)

        Parameters
        ----------
        Naverage: int
            Number of hardware averaging.
        """

        data,timestamp = self.controller.grab_spectrum()

        dwa0D_timestamp = \
            DataRaw('timestamp', units='dimensionless',
                    data=np.array([timestamp]))

        dfp = DataFromPlugins(name='Avantes', data=data, dim='Data1D',
                              labels=['data'], axes=[self.x_axis])
        self.dte_signal.emit(DataToExport(name='spectrum',
                                          data=[dfp, dwa0D_timestamp]))
 
    def stop(self):
        self.controller.abort_measurement()

    def has_dark_shutter(self):
        return False

    def has_reference_shutter(self):
        return False

if __name__ == '__main__':
    main(__file__)



import time,sys
import numpy as np
from datetime import datetime
from enum import Enum
from pymodaq_plugins_avantes.hardware import avaspec

class AvsDeviceType(Enum):

    UNKNKOWN = 0
    AS5216 = 1
    ASMINI = 2
    AS7010 = 3
    AS7007 = 4


class AvantesController:
    """
    Controller for the Avantes Spectrometer AvaSPec-ULS2048CL-EVO
    This class relies on communication with the instrument via USB-3 link

    Additional methods provide control over digital output and analog input
    See Avantes doc. and avaspec.py module
    Remark:
            - The AVS spectrometer functions are managed over 4096 pixels even
              though the spectrometer has only 2048 pixels.
            - Only a single USB spectrometer is controlled.
            - For asynchronous measurement with callback function, see either
              AvantesControllerTestApp.py or PyQt5_demo.py
    """

    def __init__(self):
        """
        Method called at object creation (implementation)
        Init object attributes
        """

        self._initialized = False
        self.hardware_present = False
        self._serial_number = str("none")
        self._measurement_config = avaspec.MeasConfigType()
        self._device_handle = 0
        self._number_of_pixels = 4096
        self._wavelengths = [0.0] * 4096
        self._spectraldata = [0.0] * 4096
        self._scan_count = 0
        # shouldn't this go to the PyMoDAQ parameters?

    def open_communication(self) -> bool:
        """
        Open the USB communication with an Avantes Spectrometer

        Returns
        -------
                True on succes, else False
        """
        # Initializes the communication interface with the spectrometers-
        # param "0" indicates USB use
        if avaspec.lib is None:
            return False

        nb_of_found_devices = avaspec.AVS_Init(0)
        nb_of_found_USB_devices = avaspec.AVS_UpdateUSBDevices()
        if nb_of_found_USB_devices < 1:
            return False

        # Get spectrometer identity, serial number and information
        device_list = avaspec.AVS_GetList(1)
        self._serial_number = str(device_list[0].SerialNumber.decode("utf-8"))
        self._initialized = True

        # Activate spectrometer for communication and get a handle on it
        self._device_handle = avaspec.AVS_Activate(device_list[0])

        # Get device configuration (number of pixels and wavelength)
        device_config = avaspec.AVS_GetParameter(self._device_handle, 63484)

        self._number_of_pixels = device_config.m_Detector_m_NrPixels
        full_wavelength_scale = avaspec.AVS_GetLambda(self._device_handle)

        # The AvaSpec_ULS2048CL_EVO spectrometer has only 2048 elements,
        # so split in 2
        self._wavelengths = np.array_split(full_wavelength_scale, 2)[0]
        #devicetype = avaspec.AVS_GetDeviceType(globals_var.dev_handle)

        self.hardware_present = True
        return True

    @property
    def is_initialized(self) -> bool:
        """
        Check if the controller is initialized

        Returns
        -------
                the initialized status True of False
        """
        return self._initialized

    @property
    def serial_number(self) -> str:
        """
        Get the serial number of the Avantes

        Returns
        -------
                serial number (str)
        """
        return self._serial_number

    def close_communication(self):
        """
        Close communication
        """
        avaspec.AVS_Done()

    def set_default_config(self):
        """
        Configure the acquisition using the _measconfig object using
        default parameters and send the configuration to the spectrometer
        """
        avaspec.AVS_UseHighResAdc(self._device_handle, True)
        # NB: return: SUCCESS = 0 or FAILURE <> 0; not currently used

        self._measurement_config.m_StartPixel = 0
        self._measurement_config.m_StopPixel = self._number_of_pixels - 1
        self._measurement_config.m_IntegrationTime = 100 # in ms
        self._measurement_config.m_IntegrationDelay = 0
        self._measurement_config.m_NrAverages = 1
        self._measurement_config.m_CorDynDark_m_Enable = 0
        # nesting of types does NOT work!! ?? whatever that means
        self._measurement_config.m_CorDynDark_m_ForgetPercentage = 0
        self._measurement_config.m_Smoothing_m_SmoothPix = 0
        self._measurement_config.m_Smoothing_m_SmoothModel = 0
        self._measurement_config.m_SaturationDetection = 0
        self._measurement_config.m_Trigger_m_Mode = 0
        self._measurement_config.m_Trigger_m_Source = 0
        self._measurement_config.m_Trigger_m_SourceType = 0
        self._measurement_config.m_Control_m_StrobeControl = 0
        self._measurement_config.m_Control_m_LaserDelay = 0
        self._measurement_config.m_Control_m_LaserWidth = 0
        self._measurement_config.m_Control_m_LaserWaveLength = 0.0
        self._measurement_config.m_Control_m_StoreToRam = 0
        result = avaspec.AVS_PrepareMeasure(self._device_handle,
                                            self._measurement_config)
        # NB: return: SUCCESS = 0 or FAILURE <> 0; not currently used

        # measurement counter
        self._scan_count = 0

        time.sleep(0.001)

    def set_integration_time(self, integration_time: float):
        """
        Set the integration time in [ms]

        Parameter
        ---------
                integration_time in [ms]
        """
        self._measurement_config.m_IntegrationTime = float(integration_time)
        avaspec.AVS_PrepareMeasure(self._device_handle,
                                   self._measurement_config)
        # NB: return: SUCCESS = 0 or FAILURE <> 0; not currently used
        self._integration_time = integration_time

    def set_number_of_averages(self, n_average: int):
        """
        Set the number of average

        Parameter
        ---------
                average_nb
        """
        self._measurement_config.m_NrAverages = n_average
        avaspec.AVS_PrepareMeasure(self._devive_handle,
                                   self._measurement_config)
        # NB: return: SUCCESS = 0 or FAILURE <> 0; not currently used

    def get_digital_input(self, pin_no: int) -> int:
        """
        Returns the status of the specified digital input

        Parameters
        ----------
            portId
            For the AS7010:
            0 = DI1 = Pin 24 at 26-pins connector
            1 = DI2 = Pin 7 at 26-pins connector
            2 = DI3 = Pin 16 at 26-pins

        Returns
        -------
                digital state 0=low or 1=high
        """

        return avaspec.AVS_GetDigIn(self._device_handle, pin_no)

    def set_digital_output(self, pin_no: int, value: int) -> int:
        """
        Set the digital output value for the specified digital output

        Parameters
        ----------
            portId
            For the AS7010:
            0 = DO1 = pin 11 at 26 - pins connector (can be used also as PWM)
            1 = DO2 = pin 2 at 26 - pins connector (can be used also as PWM)
            2 = DO3 = pin 20 at 26 - pins connector (can be used also as PWM)
            3 = DO4 = pin 12 at 26 - pins connector
            4 = DO5 = pin 3 at 26 - pins connector (can be used also as PWM)
            5 = DO6 = pin 21 at 26 - pins connector (can be used also as PWM)
            6 = DO7 = pin 13 at 26 - pins connector (can be used also as PWM)
            7 = DO8 = pin 4 at 26 - pins connector
            8 = DO9 = pin 22 at 26 - pins connector
            9 = DO10 = pin 25 at 26 - pins connector

            value: 0 (low) or 1 (high)

        Returns
        -------
                On success:    0   = ERR_SUCCESS
                On error:      -3  = ERR_DEVICE_NOT_FOUND
                               -28 = ERR_INVALID_DEVICE_ID
                               -6  = ERR_TIMEOUT (error in communication)
                               -1  = ERR_INVALID_PARAMETER
        """

        return avaspec.AVS_SetDigOut(self._device_handle, pin_no, value)

    def set_analog_output(self, pin_no: int, value: float) -> int:
        """
        Set the analog output value for the specified digital output

        Parameters
        ----------
            portId:
            For the AS7010:
            0 = AO1 = pin 17 at 26 - pins connector
            1 = AO2 = pin 26 at 26 - pins connector

            value: 0 or 5.0 V

        Returns
        -------
                On success:    0   = ERR_SUCCESS
                On error:      -3  = ERR_DEVICE_NOT_FOUND
                               -28 = ERR_INVALID_DEVICE_ID
                               -6  = ERR_TIMEOUT (error in communication)
                               -1  = ERR_INVALID_PARAMETER
        """
        return avaspec.AVS_SetAnalogOut(self._devive_handle, pin_no, value)

    @property
    def wavelengths(self):
        """
        Get an array of wavelength used by the spectrometer
        Returns
        -------
                a "standard" array of 4096 elements whom, in case of a
                Avaspec-ULS2048CL-EVO, 2048 has wavelength number,
                other elements are set to zeros  !!!
        """

        full_scale = avaspec.AVS_GetLambda(self._device_handle)
        return np.array_split(np.array(full_scale), 2)[0]

    def grab_spectrum(self):
        """
        Read the spectrum from the Avantes Spectrometer
        And store result in global_vars.spectrometer_y_values

        Returns
        -------
                none, only update globals_vars
        """
        avaspec.AVS_Measure(self._device_handle, 0, 1)
        # NB: return: SUCCESS = 0 or FAILURE <> 0; not currently used
        time.sleep(0.005)
        time.sleep(self._measurement_config.m_IntegrationTime/1000)
        # at least wait for the integration time !
        # we wait the measurement ends
        while avaspec.AVS_PollScan(self._device_handle):
            time.sleep(0.005)

        # we get the data with internal timestamp of the device
        # which is currently not used
        data = avaspec.AVS_GetScopeData(self._device_handle)
        full_spectrometer_y_values = np.array(data[1])

        # get only 2048 elements by splitting in 2
        result = np.array_split(full_spectrometer_y_values, 2)[0]
        self._scan_count += 1
        time.sleep(0.001)
        return result,data[0]

    def abort_measurement(self):
        """
        Abort a running measurement

        Returns
        -------
                SUCCESS = 0 or FAILURE <> 0
        """
        self._is_grabbing = False
        return avaspec.AVS_StopMeasure(self._device_handle)

    def start_continuous_grabbing(self, callback):
        print("controller start grabbing")
        self.pymodaq_callback = callback
        avs_cb = avaspec.AVS_MeasureCallbackFunc(self.avantes_callback)
        self._scan_count = 0
        avaspec.AVS_MeasureCallback(self._device_handle, avs_cb, -1)
        time.sleep(0.1)

    def avantes_callback(self, lparam1, lparam2):
        """
        Retrieve the spectrum from the Avantes Spectrometer
        Use this function in an asynchrone "callback" mechanism

        Returns
        -------
                spectraData: a tuple: timestamp, array of 4096 double
        """
        result = avaspec.AVS_GetScopeData(self._device_handle)
        full_data = np.array(result[1])
        self._scan_count += 1
        self.pymodaq_callback(np.array_split(full_data, 2)[0])


### simulating non existing device for debugging purpose

PIN_SIGNAL     = 2
PIN_REFERENCE  = 6
PIN_AVALIGHT   = 3
PIN_EXCITATION = 9


class AvantesSimuController(AvantesController):
    """Simulates data in case no spectrometer is connected."""

    def __init__(self, initial_line_states=0):
        AvantesController.__init__(self)
        self._pin_states = [initial_line_states for _ in range(10)]
        self._signal = 3000 * np.exp(-((self.wavelengths - 600) / 300)**4)
        self._induced = np.exp(-((self.wavelengths - 400) / 30)**2)
        self._reference = 3500 * np.exp(-((self._wavelengths - 600) / 300)**4)

    def open_communication(self):
        self._pin_states[PIN_AVALIGHT-1]   = 1
        self._pin_states[PIN_SIGNAL-1]     = 1
        self._pin_states[PIN_REFERENCE-1]  = 0
        self._pin_states[PIN_EXCITATION-1] = 1
        return True

    @staticmethod
    def close_communication():
        pass

    def configure_acquisition_by_default(self):
        pass

    def set_integration_time(self, integration_time: int):
        """
        Set the integration time in [ms]

        Parameter
        ---------
                integration_time in [ms]
        """
        self._measurement_config.m_IntegrationTime = float(integration_time)

    def set_number_of_averages(self, n_average: int):
        """
        Set the number of average

        Parameter
        ---------
                average_nb
        """
        self._measurement_config.m_NrAverages = n_average

    def get_digital_input(self, pin_no: int) -> int:
        return 0

    def set_digital_output(self, pin_no: int, value: int) -> int:
        self._pin_states[pin_no - 1] = value
        return 0

    def set_analog_output(self, pin_no: int, value: float) -> int:
        return 0

    @property
    def wavelengths(self):
        """
        Get an array of wavelength used by the spectrometer
        Returns
        -------
                a "standard" array of 4096 elements whom, in case of a
                Avaspec-ULS2048CL-EVO, 2048 has wavelength number,
                other elements are set to zeros  !!!
        """

        self._wavelengths = np.linspace(300, 900, 2048)
        return self._wavelengths

    def grab_spectrum(self):
        """
        Read the spectrum from the Avantes Spectrometer
        And store result in global_vars.spectrometer_y_values

        Returns
        -------
                none, only update globals_vars
        """
        now = time.time()
        if not hasattr(self, "start_time"):
            self.start_time = now
        drift = 1 + 0.1 * np.sin(now / 20 * 2 * np.pi)

        signal = np.random.normal(120, 4, 2048) #\
#            * (1 + 0.1 * np.sin(now / 60 * 2 * np.pi))

        if self._pin_states[PIN_AVALIGHT - 1] == 0:
            return signal, time.time()

        if self._pin_states[PIN_SIGNAL - 1] == 1:
            if self._pin_states[PIN_EXCITATION - 1] == 0:
                signal += np.random.normal(self._signal * drift, 20)
            else:
                kinetics = np.exp(-(now - self.start_time) / 120)
                signal += \
                    np.random.normal(self._signal * drift \
                                    * (1 - 0.1 * (1 - kinetics) * self._induced),
                                     20)

        if self._pin_states[PIN_REFERENCE - 1] == 1:
            signal += np.random.normal(self._reference * drift, 20)

        return signal, time.time()

    def abort_measurement(self):
        """
        Abort a running measurement

        Returns
        -------
                SUCCESS = 0 or FAILURE <> 
        """
        pass



import os.path
import sys
import inspect
import ctypes
import struct

AVS_SERIAL_LEN = 10
VERSION_LEN = 16
DETECTOR_NAME_LEN = 20
USER_ID_LEN = 64
INVALID_AVS_HANDLE_VALUE = 1000
ERR_ETHCONN_REUSE = -27
SENS_HAMS9201 = 4
SENS_TCD1304 = 5
SENS_SU256LSB = 17
SENS_SU512LDB = 18
SENS_HAMS11639 = 22
SENS_HAMG9208_512 = 24
SENS_HAMS13496 = 26
SENS_HAMS11155_2048_02_DIFF = 30
SENSOR_OFFSET = 1
NUMBER_OF_SENSOR_TYPES = 31
NR_DEFECTIVE_PIXELS = 30
MAX_NR_PIXELS = 4096
CLIENT_ID_SIZE = 32

DSTR_STATUS_DSS_MASK = 0x01   # DSTR Sequence Stop (DSS) bit of MEASUREMENT_DSTR_STATUS->DMS
DSTR_STATUS_FOE_MASK = 0x02   # FIFO Overflow Error (FOE) bit of MEASUREMENT_DSTR_STATUS->DMS
DSTR_STATUS_IERR_MASK = 0x04  # Internal Error (IERR) bit of MEASUREMENT_DSTR_STATUS->DMS

if 'linux' in sys.platform: # Linux will have 'linux' or 'linux2'
    #lib = ctypes.CDLL("/usr/local/lib/libavs.so.0")  --> modify because not the good path
    for path in ['/usr/lib64', '/usr/local/lib', '/usr/local/lib/avantes']:
        try:
            lib = ctypes.CDLL("%s/libavs.so" % path)
        except:
            pass
    func = ctypes.CFUNCTYPE
elif 'darwin' in sys.platform: # macOS will have 'darwin'
    lib = ctypes.CDLL("/usr/local/lib/libavs.0.dylib")
    func = ctypes.CFUNCTYPE
else: # Windows will have 'win32' or 'cygwin'
    import ctypes.wintypes
    if (ctypes.sizeof(ctypes.c_voidp) == 8): # 64 bit
        WM_MEAS_READY = 0x8001

        # Try to get the full path because daq_viewer with WinDLL needs a full path
        #--------------------------------------------------------------------------
        current_path = os.path.dirname(os.path.realpath(__file__))
        print(current_path)
        dll_path = os.path.join(current_path,'avaspecx64.dll')
        lib = ctypes.WinDLL(dll_path)
        # --------------------------------------------------------------------------
        #lib = ctypes.WinDLL("./avaspecx64.dll") #ok si lancement direct
        #lib = ctypes.WinDLL("c:/Users/eric/local_repository/pymodaq_plugins_avantes/src/pymodaq_plugins_avantes/hardware/avaspecx64.dll") #ligne trop longue !!!
        func = ctypes.WINFUNCTYPE
    else:
        WM_MEAS_READY = 0x0401
        lib = ctypes.WinDLL("./avaspec.dll")
        func = ctypes.WINFUNCTYPE

class AvsIdentityType(ctypes.Structure):
  _pack_ = 1
  _fields_ = [("SerialNumber", ctypes.c_char * AVS_SERIAL_LEN),
              ("UserFriendlyName", ctypes.c_char * USER_ID_LEN),
              ("Status", ctypes.c_char)]

class BroadcastAnswerType(ctypes.Structure):
  _pack_ = 1
  _fields_ = [("InterfaceType", ctypes.c_uint8),
              ("serial", ctypes.c_char * AVS_SERIAL_LEN),
              ("port", ctypes.c_uint16),
              ("status", ctypes.c_uint8),
              ("RemoteHostIp", ctypes.c_uint32),
              ("LocalIp", ctypes.c_uint32),
              ("reserved", ctypes.c_uint8 * 4)]

class MeasConfigType(ctypes.Structure):
  _pack_ = 1
  _fields_ = [("m_StartPixel", ctypes.c_uint16),
              ("m_StopPixel", ctypes.c_uint16),
              ("m_IntegrationTime", ctypes.c_float),
              ("m_IntegrationDelay", ctypes.c_uint32),
              ("m_NrAverages", ctypes.c_uint32),
              ("m_CorDynDark_m_Enable", ctypes.c_uint8), # nesting of types does NOT work!!
              ("m_CorDynDark_m_ForgetPercentage", ctypes.c_uint8),
              ("m_Smoothing_m_SmoothPix", ctypes.c_uint16),
              ("m_Smoothing_m_SmoothModel", ctypes.c_uint8),
              ("m_SaturationDetection", ctypes.c_uint8),
              ("m_Trigger_m_Mode", ctypes.c_uint8),
              ("m_Trigger_m_Source", ctypes.c_uint8),
              ("m_Trigger_m_SourceType", ctypes.c_uint8),
              ("m_Control_m_StrobeControl", ctypes.c_uint16),
              ("m_Control_m_LaserDelay", ctypes.c_uint32),
              ("m_Control_m_LaserWidth", ctypes.c_uint32),
              ("m_Control_m_LaserWaveLength", ctypes.c_float),
              ("m_Control_m_StoreToRam", ctypes.c_uint16)]

class DeviceConfigType(ctypes.Structure):
  _pack_ = 1
  _fields_ = [("m_Len", ctypes.c_uint16),
              ("m_ConfigVersion", ctypes.c_uint16),
              ("m_aUserFriendlyId", ctypes.c_char * USER_ID_LEN),
              ("m_Detector_m_SensorType", ctypes.c_uint8),                      
              ("m_Detector_m_NrPixels", ctypes.c_uint16),
              ("m_Detector_m_aFit", ctypes.c_float * 5),
              ("m_Detector_m_NLEnable", ctypes.c_bool),
              ("m_Detector_m_aNLCorrect", ctypes.c_double * 8),
              ("m_Detector_m_aLowNLCounts", ctypes.c_double),
              ("m_Detector_m_aHighNLCounts", ctypes.c_double),
              ("m_Detector_m_Gain", ctypes.c_float * 2),
              ("m_Detector_m_Reserved", ctypes.c_float),
              ("m_Detector_m_Offset", ctypes.c_float * 2),
              ("m_Detector_m_ExtOffset", ctypes.c_float),
              ("m_Detector_m_DefectivePixels", ctypes.c_uint16 * 30),
              ("m_Irradiance_m_IntensityCalib_m_Smoothing_m_SmoothPix", ctypes.c_uint16),
              ("m_Irradiance_m_IntensityCalib_m_Smoothing_m_SmoothModel", ctypes.c_uint8),
              ("m_Irradiance_m_IntensityCalib_m_CalInttime", ctypes.c_float),
              ("m_Irradiance_m_IntensityCalib_m_aCalibConvers", ctypes.c_float * 4096),
              ("m_Irradiance_m_CalibrationType", ctypes.c_uint8),
              ("m_Irradiance_m_FiberDiameter", ctypes.c_uint32),  
              ("m_Reflectance_m_Smoothing_m_SmoothPix", ctypes.c_uint16),
              ("m_Reflectance_m_Smoothing_m_SmoothModel", ctypes.c_uint8),
              ("m_Reflectance_m_CalInttime", ctypes.c_float),
              ("m_Reflectance_m_aCalibConvers", ctypes.c_float * 4096),
              ("m_SpectrumCorrect", ctypes.c_float * 4096),
              ("m_StandAlone_m_Enable", ctypes.c_bool),
              ("m_StandAlone_m_Meas_m_StartPixel", ctypes.c_uint16),
              ("m_StandAlone_m_Meas_m_StopPixel", ctypes.c_uint16),
              ("m_StandAlone_m_Meas_m_IntegrationTime", ctypes.c_float),
              ("m_StandAlone_m_Meas_m_IntegrationDelay", ctypes.c_uint32),
              ("m_StandAlone_m_Meas_m_NrAverages", ctypes.c_uint32),
              ("m_StandAlone_m_Meas_m_CorDynDark_m_Enable", ctypes.c_uint8), 
              ("m_StandAlone_m_Meas_m_CorDynDark_m_ForgetPercentage", ctypes.c_uint8),
              ("m_StandAlone_m_Meas_m_Smoothing_m_SmoothPix", ctypes.c_uint16),
              ("m_StandAlone_m_Meas_m_Smoothing_m_SmoothModel", ctypes.c_uint8),
              ("m_StandAlone_m_Meas_m_SaturationDetection", ctypes.c_uint8),
              ("m_StandAlone_m_Meas_m_Trigger_m_Mode", ctypes.c_uint8),
              ("m_StandAlone_m_Meas_m_Trigger_m_Source", ctypes.c_uint8),
              ("m_StandAlone_m_Meas_m_Trigger_m_SourceType", ctypes.c_uint8),
              ("m_StandAlone_m_Meas_m_Control_m_StrobeControl", ctypes.c_uint16),
              ("m_StandAlone_m_Meas_m_Control_m_LaserDelay", ctypes.c_uint32),
              ("m_StandAlone_m_Meas_m_Control_m_LaserWidth", ctypes.c_uint32),
              ("m_StandAlone_m_Meas_m_Control_m_LaserWaveLength", ctypes.c_float),
              ("m_StandAlone_m_Meas_m_Control_m_StoreToRam", ctypes.c_uint16),
              ("m_StandAlone_m_Nmsr", ctypes.c_int16),
              ("m_DynamicStorage", ctypes.c_uint8 * 12), # ex SD Card, do not use
              ("m_Temperature_1_m_aFit", ctypes.c_float * 5),
              ("m_Temperature_2_m_aFit", ctypes.c_float * 5),
              ("m_Temperature_3_m_aFit", ctypes.c_float * 5),
              ("m_TecControl_m_Enable", ctypes.c_bool),
              ("m_TecControl_m_Setpoint", ctypes.c_float),
              ("m_TecControl_m_aFit", ctypes.c_float * 2),
              ("m_ProcessControl_m_AnalogLow", ctypes.c_float * 2),
              ("m_ProcessControl_m_AnalogHigh", ctypes.c_float * 2),
              ("m_ProcessControl_m_DigitalLow", ctypes.c_float * 10),
              ("m_ProcessControl_m_DigitalHigh", ctypes.c_float * 10),
              ("m_EthernetSettings_m_IpAddr", ctypes.c_uint32),
              ("m_EthernetSettings_m_NetMask", ctypes.c_uint32),
              ("m_EthernetSettings_m_Gateway", ctypes.c_uint32), 
              ("m_EthernetSettings_m_DhcpEnabled", ctypes.c_uint8), 
              ("m_EthernetSettings_m_TcpPort", ctypes.c_uint16),
              ("m_EthernetSettings_m_LinkStatus", ctypes.c_uint8),
              ("m_EthernetSettings_m_ClientIdType", ctypes.c_uint8),
              ("m_EthernetSettings_m_ClientIdCustom", ctypes.c_char * 32),
              ("m_EthernetSettings_m_Reserved", ctypes.c_uint8 * 79),
              ("m_Reserved", ctypes.c_uint8 * 9608),
              ("m_OemData", ctypes.c_uint8 * 4096)]

class DstrStatusType(ctypes.Structure):
  _pack_ = 1
  _fields_ = [("m_TotalScans", ctypes.c_uint32),
              ("m_UsedScans", ctypes.c_uint32),
              ("m_Flags", ctypes.c_uint32),
              ("m_IsStopEvent", ctypes.c_uint8),
              ("m_IsOverflowEvent", ctypes.c_uint8),
              ("m_IsInternalErrorEvent", ctypes.c_uint8),
              ("m_Reserved", ctypes.c_uint8)]

def AVS_Init(a_Port = 0):
    """
    Initializes the communication interface with the spectrometers.
    
    :param a_Port: ID of port to be used, defined as follows; -1: Use both
    Ethernet(AS7010) and USB ports; 0: Use USB port; 256: Use Ethernet(AS7010)
    
    :return: Number of connected and/or found devices; ERR_CONNECTION_FAILURE,
    ERR_ETHCONN_REUSE
    """    
    prototype = func(ctypes.c_int, ctypes.c_int)
    paramflags = (1, "port",),
    AVS_Init = prototype(("AVS_Init", lib), paramflags)
    ret = AVS_Init(a_Port) 
    return ret 

def AVS_Done():
    """
    Closes the communication and releases internal storage.
    
    :return: SUCCESS = 0
    """
    prototype = func(ctypes.c_int)
    AVS_Done = prototype(("AVS_Done",lib),)
    ret = AVS_Done()
    return ret  

def AVS_GetNrOfDevices():
    """
    Deprecated function, replaced by AVS_UpdateUSBDevices(). The functionality
    is identical.
    
    :return: Number of devices found.
    """
    prototype = func(ctypes.c_int)
    AVS_GetNrOfDevices = prototype(("AVS_GetNrOfDevices", lib),)
    ret = AVS_GetNrOfDevices()
    return ret

def AVS_UpdateUSBDevices():
    """
    Internally checks the list of connected USB devices and returns the number 
    of devices attached. If AVS_Init() was called with a_Port=-1, the return 
    value also includes the number of ETH devices.
    
    :return: Number of devices found.    
    """
    prototype = func(ctypes.c_int)
    AVS_UpdateUSBDevices = prototype(("AVS_UpdateUSBDevices", lib),)
    ret = AVS_UpdateUSBDevices()
    return ret

def AVS_UpdateETHDevices(spectrometers = 1):
    """
    Returns a list containing info on all responding Ethernet spectrometers

    :param spectrometers: number of spectrometers connected. function uses 
    default value of 1, and automatically corrects.
    :return: Tuple containing BroadcastAnswerType for each found device.
    """
    prototype = func(ctypes.c_int, ctypes.c_int, ctypes.POINTER(ctypes.c_int), ctypes.POINTER(BroadcastAnswerType*spectrometers))
    paramflags = (1, "listsize",), (2, "requiredsize",), (2, "ETHlist",),
    PT_AVS_UpdateETHDevices = prototype(("AVS_UpdateETHDevices", lib), paramflags)
    reqBufferSize, ETHlist = PT_AVS_UpdateETHDevices(spectrometers*26)
    if reqBufferSize != spectrometers*26:
        ETHlist = AVS_UpdateETHDevices(reqBufferSize//26)
    return ETHlist   

def AVS_GetList(spectrometers = 1):
    """
    Returns device information for each spectrometer connected to the ports
    indicated at AVS_Init(). Wrapper function has been modified to 
    automatically update to correct listsize.
    
    :param spectrometers: number of spectrometers connected. function uses 
    default value of 1, and automatically corrects.
    :return: Tuple containing AvsIdentityType for each found device. Devices 
    are sorted by UserFriendlyName
    """
    prototype = func(ctypes.c_int, ctypes.c_int, ctypes.POINTER(ctypes.c_int), ctypes.POINTER(AvsIdentityType*spectrometers))
    paramflags = (1, "listsize",), (2, "requiredsize",), (2, "IDlist",),
    PT_GetList = prototype(("AVS_GetList", lib), paramflags)
    reqBufferSize, spectrometerList = PT_GetList(spectrometers*75)
    if reqBufferSize != spectrometers*75:
        spectrometerList = AVS_GetList(reqBufferSize//75)
    return spectrometerList

def AVS_GetHandleFromSerial(deviceSerial):
    """
    Retrieves the AvsHandle for the spectrometer with serialnumber deviceSerial. 
    Recommend usng AVS_Activate.
    
    :param deviceSerial: The serialnumber of the spectrometer
    :type deviceSerial: str, bytes
    :return: AvsHandle, handle to be used in subsequent function calls
    """
    prototype = func(ctypes.c_int, ctypes.c_char_p)
    paramflags = (1, "deviceSerial",),
    AVS_Activate = prototype(("AVS_Activate", lib), paramflags)
    if type(deviceSerial) is str:
        deviceSerial = deviceSerial.encode("utf-8")
    ret = AVS_Activate(deviceSerial)
    return ret 

def AVS_Activate(deviceId):
    """
    Activates spectrometer for communication
    
    :param deviceId: The device identifier
    :type deviceId: AvsIdentityType
    :return: AvsHandle, handle to be used in subsequent function calls
    """
    datatype = ctypes.c_byte * 75
    temp = datatype()
    x = 0
    while (x < 9): # 0 through 8
        temp[x] = deviceId.SerialNumber[x]
        x += 1
    temp[9] = 0
    x += 1    
    while (x<74): # 10 through 73
        temp[x] = 0
        x += 1
    temp[74] = int.from_bytes(deviceId.Status, byteorder='big')  #  cannot assign directly here
    prototype = func(ctypes.c_int, ctypes.c_byte * 75)
    paramflags = (1, "deviceId",),
    AVS_Activate = prototype(("AVS_Activate", lib), paramflags)
    ret = AVS_Activate(temp)
    return ret

def AVS_Deactivate(handle):
    """
    Deactivates spectrometer.
    
    :param handle: AvsHandle of the spectrometer
    :return: True when device successfully closed, False when handle not found
    """
    prototype = func(ctypes.c_bool, ctypes.c_int)
    prototype.restype = ctypes.c_bool
    paramflags = (1, "handle",),
    AVS_Deactivate = prototype(("AVS_Deactivate", lib), paramflags)
    ret = AVS_Deactivate(handle)
    return ret 

def AVS_UseHighResAdc(handle, enable):
    """
    Sets the ADC range of the spectrometer readout.
    
    :param handle: AvsHandle of the spectrometer
    :param enable: Boolean, True enables 16 bit resolution (65535 max value), 
    false uses 14 bit resolution (16383 max value)
    :return: SUCCESS = 0 or FAILURE <> 0
    """
    prototype = func(ctypes.c_int, ctypes.c_int, ctypes.c_bool)
    paramflags = (1, "handle",), (1, "enable",),
    AVS_UseHighResAdc = prototype(("AVS_UseHighResAdc", lib), paramflags)
    ret = AVS_UseHighResAdc(handle, enable)
    return ret

def AVS_GetVersionInfo(handle):
    """
    Returns three version numbers of the used system. Note that the library does 
    not check the size of the buffers allocated by the caller!
    :param handle: AvsHandle returned by AVS_Activate or others
    :return: tuple of the three requested versionstrings (FPGA, FW and Library), 
    encoded in c_char
    """       
    prototype = func(ctypes.c_int, ctypes.c_int, ctypes.c_char * VERSION_LEN, ctypes.c_char * VERSION_LEN, ctypes.c_char * VERSION_LEN)
    paramflags = (1, "handle",), (2, "FPGAversion",), (2, "FWversion",), (2, "DLLversion",),
    AVS_GetVersionInfo = prototype(("AVS_GetVersionInfo", lib), paramflags) 
    ret = AVS_GetVersionInfo(handle)
    return ret    

def AVS_PrepareMeasure(handle, measconf):
    """
    Prepares measurement on the spectrometer using the specificed configuration.
    :param handle: AvsHandle returned by AVS_Activate or others
    :param measconf: MeasConfigType containing measurement configuration.
    :return: SUCCESS = 0 or FAILURE <> 0
    """    
    prototype = func(ctypes.c_int, ctypes.c_int, ctypes.POINTER(MeasConfigType))
    paramflags = (1, "handle",), (1, "measconf",),  
    AVS_PrepareMeasure = prototype(("AVS_PrepareMeasure", lib), paramflags)
    ret = AVS_PrepareMeasure(handle, measconf)
    return ret

def AVS_Measure(handle, windowhandle, nummeas):
    """
    Starts measurement on the spectrometer, variant used for Windows messages or polling
    
    :param handle: AvsHandle of the spectrometer
    :param windowhandle: Window handle to notify application measurement result
    data is available. The library sends a Windows message to the window with 
    command WM_MEAS_READY, with SUCCESS, the number of scans that were saved in
    RAM (if enabled), or INVALID_MEAS_DATA as WPARM value and handle as LPARM 
    value. Use on Windows only, 0 to disable.
    :param nummeas: number of measurements to do. -1 is infinite, -2 is used to
    start Dynamic StoreToRam
    :return: SUCCESS = 0 or FAILURE <> 0
    """
    if not (('linux' in sys.platform) or ('darwin' in sys.platform)):
        prototype = func(ctypes.c_int, ctypes.c_int, ctypes.wintypes.HWND, ctypes.c_uint16)
    else:
        prototype = func(ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_uint16)
    paramflags = (1, "handle",), (1, "windowhandle",), (1, "nummeas"),
    AVS_Measure = prototype(("AVS_Measure", lib), paramflags)
    ret = AVS_Measure(handle, windowhandle, nummeas) 
    return ret

class AVS_MeasureCallbackFunc(object):
    def __init__(self, function):
        self.prototype = ctypes.CFUNCTYPE(None, ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int))
        self.callback = self.prototype(function)    

def AVS_MeasureCallback(handle, cb, nummeas):
    """
    Starts measurement on the spectrometer, variant used with callbacks
    
    :param handle: AvsHandle of the spectrometer
    :param cb: address of the callback function that has to be defined in the user
    program, and will be called by the library
    :param nummeas: number of measurements to do. -1 is infinite, -2 is used to
    start Dynamic StoreToRam
    :return: SUCCESS = 0 or FAILURE <> 0
    """    
    prototype = func(ctypes.c_int, ctypes.c_int, cb.prototype, ctypes.c_uint16)
    paramflags = (1, "handle",), (1, "adres",), (1, "nummeas"),
    AVS_MeasureCallback = prototype(("AVS_MeasureCallback", lib), paramflags)
    ret = AVS_MeasureCallback(handle, cb.callback, nummeas)
    return ret

class AVS_DstrCallbackFunc(object):
    def __init__(self, function):
        self.prototype = ctypes.CFUNCTYPE(None, ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_uint))
        self.callback = self.prototype(function)

def AVS_SetDstrStatusCallback(handle, cb):
    """    
    Sets the address of the callback function the library will call periodically when
    using the DynamicStoreToRam feature
    
    :param handle: AvsHandle of the spectrometer
    :param cb: address of the callback function that has to be defined in the user
    program, and will be called by the library
    :return: SUCCESS = 0 or FAILURE <> 0
    """    
    prototype = func(ctypes.c_int, ctypes.c_int, cb.prototype)
    paramflags = (1, "handle",), (1, "adres",), 
    AVS_SetDstrStatusCallback = prototype(("AVS_SetDstrStatusCallback", lib), paramflags)
    ret = AVS_SetDstrStatusCallback(handle, cb.callback)
    return ret

def AVS_GetDstrStatus(handle):
    """    
    Get the status of the buffer used in the DynamicStoreToRam feature
    
    :param handle: AvsHandle of the spectrometer
    :return: DstrStatusType
    """      
    prototype = func(ctypes.c_int, ctypes.c_int, ctypes.POINTER(DstrStatusType))
    paramflags = (1, "handle",), (2, "dstrstatus",),
    AVS_GetDstrStatus = prototype(("AVS_GetDstrStatus", lib), paramflags)
    ret = AVS_GetDstrStatus(handle)
    return ret

def AVS_StopMeasure(handle):
    """    
    Stops a running measurement
    
    :param handle: AvsHandle of the spectrometer
    :return: SUCCESS = 0 or FAILURE <> 0
    """      
    prototype = func(ctypes.c_int, ctypes.c_int)
    paramflags = (1, "handle",),
    AVS_StopMeasure = prototype(("AVS_StopMeasure", lib), paramflags)
    ret = AVS_StopMeasure(handle)
    return ret

def AVS_PollScan(handle):
    """    
    Will show whether new measurement data are available
    
    :param handle: AvsHandle of the spectrometer
    :return: 0 = no data available or 1 = data available
    """  
    prototype = func(ctypes.c_bool, ctypes.c_int)
    paramflags = (1, "handle",),
    AVS_PollScan = prototype(("AVS_PollScan", lib), paramflags)
    ret = AVS_PollScan(handle)
    return ret
    
def AVS_GetScopeData(handle):
    """
    Returns the pixel values of the last performed measurement. Should be 
    called after the notification on AVS_Measure is triggered. 
    
    :param handle: the AvsHandle of the spectrometer
    :return timestamp: ticks count last pixel of spectrum is received by 
    microcontroller ticks in 10 microsecond units since spectrometer started
    :return spectrum: 4096 element array of doubles, pixels values of spectrometer
    """
    prototype = func(ctypes.c_int, ctypes.c_int, ctypes.POINTER(ctypes.c_uint32), ctypes.POINTER(ctypes.c_double * 4096))
    paramflags = (1, "handle",), (2, "timelabel",), (2, "spectrum",),
    AVS_GetScopeData = prototype(("AVS_GetScopeData", lib), paramflags)
    timestamp, spectrum = AVS_GetScopeData(handle)
    return timestamp, spectrum

def AVS_GetSaturatedPixels(handle):
    """
    Returns the saturation values of the last performed measurement. Should be 
    called after AVS_GetScopeData. 
    
    :param handle: the AvsHandle of the spectrometer
    :return saturated: 4096 element array of bytes, 1 = saturated and 0 = not saturated
    """
    prototype = func(ctypes.c_int, ctypes.c_int, ctypes.POINTER(ctypes.c_uint8 * 4096))    
    paramflags = (1, "handle",), (2, "saturated",),
    AVS_GetSaturatedPixels = prototype(("AVS_GetSaturatedPixels", lib), paramflags)
    saturated = AVS_GetSaturatedPixels(handle)
    return saturated 

def AVS_GetLambda(handle):
    """
    Returns the wavelength values corresponding to the pixels if available. 
    This information is stored in the Library during the AVS_Activate() procedure.
    
    :param handle: the AvsHandle of the spectrometer
    :return: 4096 element array of wavelength values for pixels. If the detector
    is less than 4096 pixels, zeros are returned for extra pixels.
    """
    prototype = func(ctypes.c_int, ctypes.c_int, ctypes.POINTER(ctypes.c_double * 4096))
    paramflags = (1, "handle",), (2, "wavelength",),
    AVS_GetLambda = prototype(("AVS_GetLambda", lib), paramflags)
    ret = AVS_GetLambda(handle)
    return ret

def AVS_GetNumPixels(handle):
    """
    Returns the number of pixels of a spectrometer. This information is stored 
    in the Library during the AVS_Activate() procedure.
    
    :param handle: the AvsHandle of the spectrometer
    :return: unsigned integer, number of pixels in spectrometer
    """
    prototype = func(ctypes.c_int, ctypes.c_int, ctypes.POINTER(ctypes.c_short))
    paramflags = (1, "handle",), (2, "numPixels",),
    AVS_GetNumPixels = prototype(("AVS_GetNumPixels",lib), paramflags)
    ret = AVS_GetNumPixels(handle)
    return ret    

def AVS_GetDigIn(handle, portId):
    """
    Returns the status of the specified digital input.
    
    :param handle: the AvsHandle of the spectrometer
    :param portId: the identifier of the digital input 
    :return: the value of the digital input, 0 = low and 1 = high
    """    
    prototype = func(ctypes.c_int, ctypes.c_int, ctypes.c_uint8, ctypes.POINTER(ctypes.c_uint8))
    paramflags = (1, "handle",), (1, "portId",), (2, "value",),
    AVS_GetDigIn = prototype(("AVS_GetDigIn", lib), paramflags)
    ret = AVS_GetDigIn(handle, portId) 
    return ret

def AVS_SetDigOut(handle, portId, value):
    """
    Sets the status of the specified digital output.
    
    :param handle: the AvsHandle of the spectrometer
    :param portId: the identifier of the digital output
    :param value: the value of the digital output, 0 = low and 1 = high 
    :return: SUCCESS = 0 or FAILURE <> 0 
    """       
    prototype = func(ctypes.c_int, ctypes.c_int, ctypes.c_uint8, ctypes.c_uint8)
    paramflags = (1, "handle",), (1, "portId",), (1, "value",),
    AVS_SetDigOut = prototype(("AVS_SetDigOut", lib), paramflags)
    ret = AVS_SetDigOut(handle, portId, value)
    return ret

def AVS_SetPwmOut(handle, portId, frequency, dutycycle):
    """
    Selects the PWM functionality for the specified digital output.
    
    :param handle: the AvsHandle of the spectrometer
    :param portId: the identifier of the digital output
    :param frequency: the desired PWM frequency (500 - 300000 Hz)
    :param dutycycle: the percentage high time in one cycle (0-100)
    :return: SUCCESS = 0 or FAILURE <> 0 
    """       
    prototype = func(ctypes.c_int, ctypes.c_int, ctypes.c_uint8, ctypes.c_uint32, ctypes.c_uint8)
    paramflags = (1, "handle",), (1, "portId",), (1, "frequency",), (1, "dutycycle",),
    AVS_SetPwmOut = prototype(("AVS_SetPwmOut", lib), paramflags)
    ret = AVS_SetPwmOut(handle, portId, frequency, dutycycle)
    return ret    

def AVS_GetAnalogIn(handle, portId):
    """
    Returns the status of the specified analog input.
    
    :param handle: the AvsHandle of the spectrometer
    :param portId: the identifier of the analog input 
    :return: the value of the analog input, in Volts (or degrees Celsius)
    """      
    prototype = func(ctypes.c_int, ctypes.c_int, ctypes.c_uint8, ctypes.POINTER(ctypes.c_float))
    paramflags = (1, "handle",), (1, "portId",), (2, "value",),
    AVS_GetAnalogIn = prototype(("AVS_GetAnalogIn", lib), paramflags)
    ret = AVS_GetAnalogIn(handle, portId)
    return ret

def AVS_SetAnalogOut(handle, portId, value):
    """
    Sets the analog output value for the specified analog output.
    
    :param handle: the AvsHandle of the spectrometer
    :param portId: the identifier of the analog output
    :param value: the value of the analog output in Volts (0 - 5.0V) 
    :return: SUCCESS = 0 or FAILURE <> 0 
    """      
    prototype = func(ctypes.c_int, ctypes.c_int, ctypes.c_uint8, ctypes.c_float)
    paramflags = (1, "handle",), (1, "portId",), (1, "value",),
    AVS_SetAnalogOut = prototype(("AVS_SetAnalogOut", lib), paramflags)
    ret = AVS_SetAnalogOut(handle, portId, value)
    return ret

def AVS_GetParameter(handle, size = 63484):
    """
    Returns the device information of the spectrometer.
    
    :param handle: the AvsHandle of the spectrometer
    :param size: size in bytes allocated to store DeviceConfigType
    :return: DeviceConfigType structure containing spectrometer configuration data
    """
    prototype = func(ctypes.c_int, ctypes.c_int, ctypes.c_uint32, ctypes.POINTER(ctypes.c_uint32), ctypes.POINTER(DeviceConfigType))
    paramflags = (1, "handle",), (1, "size",), (2, "reqsize",), (2, "deviceconfig",),
    AVS_GetParameter = prototype(("AVS_GetParameter", lib), paramflags)
    ret = AVS_GetParameter(handle, size)
    if ret[0] != size:
        ret = AVS_GetParameter(ret[0])
    return ret[1]

def AVS_SetParameter(handle, deviceconfig):
    """
    Overwrites the device information of the spectrometer with the specified values.
    The data is not checked. Use with care!
    
    :param handle: the AvsHandle of the spectrometer
    :param deviceconfig: the DeviceConfigType structure that will be sent to the spectrometer
    :return: SUCCESS = 0 or FAILURE <> 0 
    """   
    prototype = func(ctypes.c_int, ctypes.c_int, ctypes.POINTER(DeviceConfigType))
    paramflags = (1, "handle",), (1, "deviceconfig",),  
    AVS_SetParameter = prototype(("AVS_SetParameter", lib), paramflags)
    ret = AVS_SetParameter(handle, deviceconfig)
    return ret

def AVS_ResetParameter(handle):
    """
    Resets the device information of the spectrometer to the factory defaults.
    This will result in the loss of all user changes made using AVS_SetParameter
    
    :param handle: the AvsHandle of the spectrometer
    :return: SUCCESS = 0 or FAILURE <> 0 
    """       
    prototype = func(ctypes.c_int, ctypes.c_int)
    paramflags = (1, "handle",),
    AVS_ResetParameter = prototype(("AVS_ResetParameter", lib), paramflags)
    ret = AVS_ResetParameter(handle)
    return ret 

def AVS_SetSyncMode(handle, enable):
    """
    Disables/Enables support for synchronous measurement. Library takes care of 
    dividing Nmsr request into Nmsr number of single measurement requests.
    
    See AvaSpec Library Manual section 3.4.8 for more information on running 
    multiple spectrometers synchronized.
    
    :param handle: AvsHandle of the master device spectrometer.
    :param enable: Boolean, 0 disables sync mode, 1 enables sync mode
    :return: SUCCESS = 0 or FAILURE <> 0 
    """
    prototype = func(ctypes.c_int, ctypes.c_int, ctypes.c_bool)
    paramflags = (1, "handle",), (1, "enable",),
    AVS_SetSyncMode = prototype(("AVS_SetSyncMode", lib), paramflags)
    ret = AVS_SetSyncMode(handle, enable)
    return ret

def AVS_GetDeviceType(handle):
    """
    Returns the type of the spectrometer, defined by its PCB
    
    :param handle: the AvsHandle of the spectrometer
    :return: integer value, 0=unknown, 1=AS5216, 2=ASMINI, 3=AS7010
    """
    prototype = func(ctypes.c_int, ctypes.c_int, ctypes.POINTER(ctypes.c_byte))
    paramflags = (1, "handle",), (2, "devicetype",),
    AVS_GetDeviceType = prototype(("AVS_GetDeviceType",lib), paramflags)
    ret = AVS_GetDeviceType(handle)
    return ret 

def AVS_GetDetectorName(handle, SensorType):
    """
    Returns the name of the detector inside the spectrometer.
    
    :param handle: the AvsHandle of the spectrometer
    :param Sensortype: byte value that defines the detector type, part of the Device Configuration
    :return: Detector name, encoded in c_char, a null terminated string
    """
    prototype = func(ctypes.c_int, ctypes.c_int, ctypes.c_byte, ctypes.c_char * DETECTOR_NAME_LEN)
    paramflags = (1, "handle",), (1, "SensorType",), (2, "SensorName",), 
    AVS_GetDetectorName = prototype(("AVS_GetDetectorName", lib), paramflags) 
    ret = AVS_GetDetectorName(handle, SensorType)
    return ret 

def AVS_SetSensitivityMode(handle, enable):
    """
    Selects between LowNoise and HighSensitivity mode for certain detectors.
      
    :param handle: AvsHandle of the spectrometer.
    :param enable: unsigned integer, 0 sets LowNoise mode, 1 sets HighSensitivity mode 
    """
    prototype = func(ctypes.c_int, ctypes.c_int, ctypes.c_uint32)
    paramflags = (1, "handle",), (1, "enable",),
    AVS_SetSensitivityMode = prototype(("AVS_SetSensitivityMode", lib), paramflags)
    ret = AVS_SetSensitivityMode(handle, enable)
    return ret

def AVS_SetPrescanMode(handle, enable):
    """
    Selects between PreScan and ClearBuffer mode for the Toshiba 3648 detector.
      
    :param handle: AvsHandle of the spectrometer.
    :param enable: boolean, 0 sets ClearBuffer mode, 1 sets PreScan mode (default mode)
    """    
    prototype = func(ctypes.c_int, ctypes.c_int, ctypes.c_bool)
    paramflags = (1, "handle",), (1, "enable",),
    AVS_SetPrescanMode = prototype(("AVS_SetPrescanMode", lib), paramflags)
    ret = AVS_SetPrescanMode(handle, enable)
    return ret

def AVS_ResetDevice(handle):
    """
    Performs a hard reset on the given spectrometer.
      
    :param handle: AvsHandle of the spectrometer.
    :return: SUCCESS = 0 or FAILURE <> 0
    """     
    prototype = func(ctypes.c_int, ctypes.c_int)
    paramflags = (1, "handle",),
    AVS_ResetDevice = prototype(("AVS_ResetDevice", lib), paramflags)
    ret = AVS_ResetDevice(handle)
    return ret

def AVS_EnableLogging(enable):
    """
    Enables or disables writing debug information to a log file, called "avaspec.dll.log", located in your user directory.
    Implemented for Windows only.
    
    :param enable: Boolean, True enables logging, False disables logging
    :return: True = 1
    """    
    prototype = func(ctypes.c_int, ctypes.c_bool)
    paramflags = (1, "enable",),
    AVS_EnableLogging = prototype(("AVS_EnableLogging", lib), paramflags)
    ret = AVS_EnableLogging(enable)    
    return ret    
